import numpy as np
import pandas as pd
from libpysal import weights
from sklearn.base import BaseEstimator

from esda.crand import _prepare_univariate
from esda.crand import crand as _crand_plus
from esda.crand import njit as _njit

PERMUTATIONS = 999


class Join_Counts_Local_MV(BaseEstimator):  # noqa: N801
    """Multivariate Local Join Count Statistic"""

    def __init__(
        self,
        connectivity=None,
        permutations=PERMUTATIONS,
        n_jobs=1,
        keep_simulations=True,
        seed=None,
        island_weight=0,
        drop_islands=True,
    ):
        """
        Initialize a Local_Join_Counts_MV estimator

        Parameters
        ----------
        connectivity : W | Graph
            spatial weights instance as W or Graph aligned with y
        permutations : int
            number of random permutations for calculation of pseudo p_values
        n_jobs : int
            Number of cores to be used in the conditional randomisation. If -1,
            all available cores are used.
        keep_simulations : bool (default True)
           If ``True``, the entire matrix of replications under the null is stored
           in memory and accessible; otherwise, replications are not saved
        seed : int (default None)
            Seed to ensure reproducibility of conditional randomizations.
            Must be set here, and not outside of the function, since numba
            does not correctly interpret external seeds
            nor numpy.random.RandomState instances.
        island_weight : int or float (default 0)
            value to use as a weight for the "fake" neighbor for every island.
            If ``numpy.nan``, will propagate to the final local statistic depending
            on the ``stat_func``. If ``0``, then the lag is always zero for islands.
        drop_islands : bool (default True)
            Whether or not to preserve islands as entries in the adjacency
            list. By default, observations with no neighbors do not appear
            in the adjacency list. If islands are kept, they are coded as
            self-neighbors with zero weight. See ``libpysal.weights.to_adjlist()``.
        """

        self.connectivity = connectivity
        self.permutations = permutations
        self.n_jobs = n_jobs
        self.keep_simulations = keep_simulations
        self.seed = seed
        self.island_weight = island_weight
        self.drop_islands = drop_islands

    def fit(self, variables, n_jobs=1, permutations=999):
        """
        Parameters
        ----------
        variables     : numpy.ndarray
                        array(s) containing binary (0/1) data
        Returns
        -------
        the fitted estimator.

        Notes
        -----
        Technical details and derivations can be found in :cite:`AnselinLi2019`.

        Examples
        --------
        >>> import libpysal
        >>> w = libpysal.weights.lat2W(4, 4)
        >>> x = np.ones(16)
        >>> x[0:8] = 0
        >>> z = [0,1,0,1,1,1,1,1,0,0,1,1,0,0,1,1]
        >>> y = [0,1,1,1,1,1,1,1,0,0,0,1,0,0,1,1]
        >>> LJC_MV = Local_Join_Counts_MV(connectivity=w).fit([x, y, z])
        >>> LJC_MV.LJC
        >>> LJC_MV.p_sim

        Guerry data extending GeoDa tutorial
        >>> import libpysal
        >>> import geopandas as gpd
        >>> guerry = libpysal.examples.load_example('Guerry')
        >>> guerry_ds = gpd.read_file(guerry.get_path('Guerry.shp'))
        >>> guerry_ds['infq5'] = 0
        >>> guerry_ds['donq5'] = 0
        >>> guerry_ds['suic5'] = 0
        >>> guerry_ds.loc[(guerry_ds['Infants'] > 23574), 'infq5'] = 1
        >>> guerry_ds.loc[(guerry_ds['Donatns'] > 10973), 'donq5'] = 1
        >>> guerry_ds.loc[(guerry_ds['Suicids'] > 55564), 'suic5'] = 1
        >>> w = libpysal.weights.Queen.from_dataframe(guerry_ds)
        >>> LJC_MV = Local_Join_Counts_MV(
        ...     connectivity=w
        ... ).fit([guerry_ds['infq5'], guerry_ds['donq5'], guerry_ds['suic5']])
        >>> LJC_MV.LJC
        >>> LJC_MV.p_sim
        """

        w = self.connectivity
        # Fill the diagonal with 0s
        if isinstance(w, weights.W):
            w = weights.util.fill_diagonal(w, val=0)
            w.transform = "b"
        else:
            w = w.assign_self_weight(0).eliminate_zeros().transform("b")

        self.n = len(variables[0])
        self.w = w

        self.variables = np.array(variables, dtype="float")

        # Need to ensure that the product is an
        # np.array() of dtype='float' for numba
        self.ext = np.array(np.prod(np.vstack(variables), axis=0), dtype="float")

        self.LJC = self._statistic(variables, w, self.drop_islands)

        if permutations:
            self.p_sim, self.rjoins = _crand_plus(
                z=self.ext,
                w=self.w,
                observed=self.LJC,
                permutations=permutations,
                keep=True,
                n_jobs=n_jobs,
                stat_func=_ljc_mv,
                island_weight=self.island_weight,
            )
            # Set p-values for those with LJC of 0 to NaN
            self.p_sim[self.LJC == 0] = "NaN"

        return self

    @staticmethod
    def _statistic(variables, w, drop_islands):
        # Create adjacency list. Note that remove_symmetric=False -
        # different from the esda.Join_Counts() function.
        if isinstance(w, weights.W):
            adj_list = w.to_adjlist(remove_symmetric=False, drop_islands=drop_islands)
            # The zseries
            zseries = [pd.Series(i, index=w.id_order) for i in variables]
        else:
            if drop_islands:
                adj_list = w.adjacency.drop(w.isolates).reset_index()
            else:
                adj_list = w.adjacency.reset_index()
            zseries = [pd.Series(i, index=w.unique_ids) for i in variables]
        # The focal values
        focal = [zseries[i].loc[adj_list.focal].values for i in range(len(variables))]
        # The neighbor values
        neighbor = [
            zseries[i].loc[adj_list.neighbor].values for i in range(len(variables))
        ]

        # Find instances where all surrounding
        # focal and neighbor values == 1
        focal_all = np.array(np.all(np.dstack(focal) == 1, axis=2))
        neighbor_all = np.array(np.all(np.dstack(neighbor) == 1, axis=2))
        MCLC = (focal_all) & (neighbor_all)
        # Convert list of True/False to boolean array
        # and unlist (necessary for building pd.DF)
        MCLC = list(MCLC * 1)

        # Create a df that uses the adjacency list
        # focal values and the BBs counts
        adj_list_MCLC = pd.DataFrame(adj_list.focal.values, MCLC).reset_index()
        # Temporarily rename the columns
        adj_list_MCLC.columns = ["MCLC", "ID"]
        adj_list_MCLC = adj_list_MCLC.groupby(by="ID").sum()

        return np.array(adj_list_MCLC.MCLC.values, dtype="float")


# --------------------------------------------------------------
# Conditional Randomization Function Implementations
# --------------------------------------------------------------

# Note: scaling not used


@_njit(fastmath=True)
def _ljc_mv(i, z, permuted_ids, weights_i, scaling):  # noqa: ARG001
    other_weights = weights_i[1:]
    zi, zrand = _prepare_univariate(i, z, permuted_ids, other_weights)
    return zi * (zrand @ other_weights)



import unittest

import geopandas as gpd
import libpysal
import numpy as np
import pandas as pd
import pytest
from libpysal.common import ATOL, RTOL
from numpy.testing import assert_array_equal
from packaging.version import Version

from .. import moran

SEED = 12345
GPD_GE_120 = (Version(gpd.__version__) >= Version("1.1.2.dev")) and Version(
    gpd.__version__
).is_devrelease

parametrize_stl = pytest.mark.parametrize(
    "w",
    [
        libpysal.io.open(libpysal.examples.get_path("stl.gal")).read(),
        libpysal.graph.Graph.from_W(
            libpysal.io.open(libpysal.examples.get_path("stl.gal")).read()
        ),
    ],
    ids=["W", "Graph"],
)
parametrize_sids = pytest.mark.parametrize(
    "w",
    [
        libpysal.io.open(libpysal.examples.get_path("sids2.gal")).read(),
        libpysal.graph.Graph.from_W(
            libpysal.io.open(libpysal.examples.get_path("sids2.gal")).read()
        ),
    ],
    ids=["W", "Graph"],
)

parametrize_lat3x3 = pytest.mark.parametrize(
    "w",
    [
        libpysal.weights.util.lat2W(3, 3),
        libpysal.graph.Graph.from_W(libpysal.weights.util.lat2W(3, 3)),
    ],
    ids=["W", "Graph"],
)
parametrize_desmith = pytest.mark.parametrize(
    "w",
    [
        libpysal.io.open(libpysal.examples.get_path("desmith.gal")).read(),
        libpysal.graph.Graph.from_W(
            libpysal.io.open(libpysal.examples.get_path("desmith.gal")).read()
        ),
    ],
    ids=["W", "Graph"],
)

sac1 = libpysal.examples.load_example("Sacramento1")
sac1 = gpd.read_file(sac1.get_path("sacramentot2.shp"))

parametrize_sac = pytest.mark.parametrize(
    "w",
    [
        libpysal.weights.Queen.from_dataframe(sac1),
        libpysal.graph.Graph.build_contiguity(sac1, rook=False),
    ],
    ids=["W", "Graph"],
)


class TestMoran:
    def setup_method(self):
        f = libpysal.io.open(libpysal.examples.get_path("stl_hom.txt"))
        self.y = np.array(f.by_col["HR8893"])

    @parametrize_stl
    def test_moran(self, w):
        mi = moran.Moran(self.y, w, two_tailed=False)
        np.testing.assert_allclose(mi.I, 0.24365582621771659, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(mi.p_norm, 0.00013573931385468807)

    @parametrize_sids
    def test_sids(self, w):
        f = libpysal.io.open(libpysal.examples.get_path("sids2.dbf"))
        SIDR = np.array(f.by_col("SIDR74"))
        mi = moran.Moran(SIDR, w, two_tailed=False)
        np.testing.assert_allclose(mi.I, 0.24772519320480135, atol=ATOL, rtol=RTOL)
        np.testing.assert_allclose(mi.p_norm, 5.7916539074498452e-05)

    @parametrize_lat3x3
    def test_variance(self, w):
        y = np.arange(1, 10)
        mi = moran.Moran(y, w, transformation="B")
        np.testing.assert_allclose(
            mi.VI_rand, 0.059687500000000004, atol=ATOL, rtol=RTOL
        )
        np.testing.assert_allclose(
            mi.VI_norm, 0.053125000000000006, atol=ATOL, rtol=RTOL
        )

    @parametrize_stl
    def test_z_consistency(self, w):
        m1 = moran.Moran(self.y, w)
        # m2 = moran.Moran_BV(self.x, self.y, self.w) TODO testing for other.z values
        m3 = moran.Moran_Local(self.y, w, keep_simulations=True, seed=SEED)
        # m4 = moran.Moran_Local_BV(self.x, self.y, self.w)
        np.testing.assert_allclose(m1.z, m3.z, atol=ATOL, rtol=RTOL)

    @pytest.mark.skip("This function is being deprecated in the next release.")
    def test_by_col(self):
        from libpysal.io import geotable as pdio

        df = pdio.read_files(libpysal.examples.get_path("sids2.dbf"))
        w = libpysal.io.open(libpysal.examples.get_path("sids2.gal")).read()
        mi = moran.Moran.by_col(df, ["SIDR74"], w=w, two_tailed=False)
        sidr = np.unique(mi.SIDR74_moran.values).item()
        pval = np.unique(mi.SIDR74_p_sim.values).item()
        np.testing.assert_allclose(sidr, 0.24772519320480135, atol=ATOL, rtol=RTOL)
        np.testing.assert_allclose(pval, 0.001)

    @parametrize_sac
    def test_plot_simulation(self, w):
        pytest.importorskip("seaborn")

        m = moran.Moran(sac1.WHITE, w=w)
        ax = m.plot_simulation()

        assert len(ax.collections) == 3

        kde = ax.collections[0]
        np.testing.assert_array_almost_equal(
            kde.get_facecolor(),
            [[0.7294117647058823, 0.7294117647058823, 0.7294117647058823, 0.25]],
        )
        assert kde.get_fill()
        assert len(kde.get_paths()[0]) == 403

        i_vline = ax.collections[1]
        np.testing.assert_array_almost_equal(
            i_vline.get_color(),
            [[0.8392156862745098, 0.3764705882352941, 0.30196078431372547, 1.0]],
        )
        assert i_vline.get_label() == "Moran's I"
        np.testing.assert_array_almost_equal(
            i_vline.get_paths()[0].vertices,
            np.array([[m.I, 0.0], [m.I, 1.0]]),
        )

        ei_vline = ax.collections[2]
        np.testing.assert_array_almost_equal(
            ei_vline.get_color(),
            [[0.12156863, 0.46666667, 0.70588235, 1.0]],
        )
        assert ei_vline.get_label() == "Expected I"
        np.testing.assert_array_almost_equal(
            ei_vline.get_paths()[0].vertices,
            np.array([[m.EI, 0.0], [m.EI, 1.0]]),
        )

    @parametrize_sac
    def test_plot_simulation_custom(self, w):
        pytest.importorskip("seaborn")
        plt = pytest.importorskip("matplotlib.pyplot")

        m = moran.Moran(sac1.WHITE, w=w)

        _, ax = plt.subplots(figsize=(12, 12))
        ax = m.plot_simulation(
            ax=ax, fitline_kwds={"color": "red"}, color="pink", shade=False, legend=True
        )

        assert len(ax.collections) == 2
        assert len(ax.lines) == 1

        kde = ax.lines[0]
        np.testing.assert_array_almost_equal(
            kde.get_color(),
            [1.0, 0.75294118, 0.79607843, 1],
        )
        assert len(kde.get_path()) == 200

        i_vline = ax.collections[0]
        np.testing.assert_array_almost_equal(
            i_vline.get_color(),
            [[1.0, 0.0, 0.0, 1.0]],
        )
        assert i_vline.get_label() == "Moran's I"
        np.testing.assert_array_almost_equal(
            i_vline.get_paths()[0].vertices,
            np.array([[m.I, 0.0], [m.I, 1.0]]),
        )

        ei_vline = ax.collections[1]
        np.testing.assert_array_almost_equal(
            ei_vline.get_color(),
            [[0.12156863, 0.46666667, 0.70588235, 1.0]],
        )
        assert ei_vline.get_label() == "Expected I"
        np.testing.assert_array_almost_equal(
            ei_vline.get_paths()[0].vertices,
            np.array([[m.EI, 0.0], [m.EI, 1.0]]),
        )

        assert ax.get_legend_handles_labels()[1] == [
            "Distribution of simulated Is",
            "Moran's I",
            "Expected I",
        ]

    @parametrize_sac
    def test_plot_scatter(self, w):
        import matplotlib

        matplotlib.use("Agg")

        m = moran.Moran(
            sac1.WHITE,
            w,
        )

        ax = m.plot_scatter()

        # test scatter
        np.testing.assert_array_almost_equal(
            ax.collections[0].get_facecolors(),
            np.array([[0.729412, 0.729412, 0.729412, 0.6]]),
        )

        # test fitline
        l_ = ax.lines[2]
        x, y = l_.get_data()
        np.testing.assert_almost_equal(x.min(), -1.8236414387225368)
        np.testing.assert_almost_equal(x.max(), 3.893056527659032)
        np.testing.assert_almost_equal(y.min(), -0.7371749399524187)
        np.testing.assert_almost_equal(y.max(), 1.634939204358587)
        assert l_.get_color() == "#d6604d"

    @parametrize_sac
    def test_plot_scatter_args(self, w):
        import matplotlib

        matplotlib.use("Agg")

        m = moran.Moran(
            sac1.WHITE,
            w,
        )

        ax = m.plot_scatter(
            scatter_kwds=dict(color="blue"), fitline_kwds=dict(color="pink")
        )

        # test scatter
        np.testing.assert_array_almost_equal(
            ax.collections[0].get_facecolors(),
            np.array([[0, 0, 1, 0.6]]),
        )

        # test fitline
        l_ = ax.lines[2]
        assert l_.get_color() == "pink"


class TestMoranRate:
    def setup_method(self):
        f = libpysal.io.open(libpysal.examples.get_path("sids2.dbf"))
        self.e = np.array(f.by_col["SID79"])
        self.b = np.array(f.by_col["BIR79"])

    @parametrize_sids
    def test_moran_rate(self, w):
        mi = moran.Moran_Rate(self.e, self.b, w, two_tailed=False)
        np.testing.assert_allclose(mi.I, 0.16622343552567395, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(mi.p_norm, 0.004191499504892171)

    @pytest.mark.skip("This function is being deprecated in the next release.")
    def test_by_col(self):
        from libpysal.io import geotable as pdio

        df = pdio.read_files(libpysal.examples.get_path("sids2.dbf"))
        mi = moran.Moran_Rate.by_col(
            df, ["SID79"], ["BIR79"], w=self.w, two_tailed=False
        )
        sidr = np.unique(mi["SID79-BIR79_moran_rate"].values).item()
        pval = np.unique(mi["SID79-BIR79_p_sim"].values).item()
        np.testing.assert_allclose(sidr, 0.16622343552567395, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(pval, 0.008)


class TestMoranBVmatrix:
    def setup_method(self):
        f = libpysal.io.open(libpysal.examples.get_path("sids2.dbf"))
        varnames = ["SIDR74", "SIDR79", "NWR74", "NWR79"]
        self.names = varnames
        vars_ = [np.array(f.by_col[var]) for var in varnames]
        self.vars_ = vars_

    @parametrize_sids
    def test_defaults(self, w):
        res = moran.Moran_BV_matrix(self.vars_, w, varnames=self.names)
        np.testing.assert_allclose(res[(0, 1)].I, 0.19362610652874668)
        np.testing.assert_allclose(res[(3, 0)].I, 0.37701382542927858)

    @parametrize_sids
    def test_plot_moran_facet(self, w):
        matrix = moran.Moran_BV_matrix(self.vars_, w, varnames=self.names)
        axes = moran.plot_moran_facet(matrix)
        assert axes.shape == (4, 4)

        assert axes[0][0].spines["left"].get_visible()
        assert not axes[0][0].spines["bottom"].get_visible()
        assert axes[3][0].spines["left"].get_visible()
        assert axes[3][0].spines["bottom"].get_visible()
        assert not axes[3][1].spines["left"].get_visible()
        assert axes[3][1].spines["bottom"].get_visible()
        assert not axes[1][1].spines["left"].get_visible()
        assert not axes[1][1].spines["bottom"].get_visible()

        np.testing.assert_array_almost_equal(
            axes[1][1].get_facecolor(),
            (0.8509803921568627, 0.8509803921568627, 0.8509803921568627, 1.0),
        )


class TestMoranLocal:
    def setup_method(self):
        f = libpysal.io.open(libpysal.examples.get_path("desmith.txt"))
        self.y = np.array(f.by_col["z"])

    @parametrize_desmith
    def test_defaults(self, w):
        lm = moran.Moran_Local(
            self.y,
            w,
            transformation="r",
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )
        np.testing.assert_allclose(lm.z_sim[0], -0.6990291160835514)
        np.testing.assert_allclose(lm.p_z_sim[0], 0.24226691753791396)

    @parametrize_sac
    def test_labels(self, w):
        lm = moran.Moran_Local(
            sac1.HSG_VAL.values,
            w,
            transformation="r",
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )
        expected_labels = np.array(
            [
                "High-High",
                "High-High",
                "Insignificant",
                "High-High",
                "Insignificant",
                "High-High",
                "High-High",
                "High-High",
                "Insignificant",
                "Insignificant",
            ]
        )
        assert_array_equal(lm.get_cluster_labels()[:10], expected_labels)
        assert_array_equal(
            pd.Series(lm.get_cluster_labels(0.05)).value_counts().values,
            np.array([277, 82, 38, 3, 3]),
        )

    @parametrize_sac
    def test_explore(self, w):
        lm = moran.Moran_Local(
            sac1.HSG_VAL.values,
            w,
            transformation="r",
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )
        m = lm.explore(sac1)
        np.testing.assert_array_equal(
            m.get_bounds(),
            [[38.018422, -122.422049], [39.316476, -119.877249]],
        )
        assert len(m.to_dict()["children"]) == 3

        out_str = _fetch_map_string(m)

        assert '"High-High","__folium_color":"#d7191c"' in out_str
        assert '"Low-High","__folium_color":"#89cff0"' in out_str
        assert '"Low-Low","__folium_color":"#2c7bb6"' in out_str
        assert '"High-Low","__folium_color":"#fdae61"' in out_str
        assert '"Insignificant","__folium_color":"#d3d3d3"' in out_str

        assert out_str.count("#d7191c") == 41
        assert out_str.count("#89cff0") == 6
        assert out_str.count("#2c7bb6") == 85
        assert out_str.count("#fdae61") == 6
        assert out_str.count("#d3d3d3") == 280

    @parametrize_sac
    def test_plot(self, w):
        import matplotlib

        matplotlib.use("Agg")

        lm = moran.Moran_Local(
            sac1.HSG_VAL.values,
            w,
            transformation="r",
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )
        ax = lm.plot(sac1)

        if GPD_GE_120:
            expected = [
                [[0.84313725, 0.09803922, 0.10980392, 1.0]],
                [[0.99215686, 0.68235294, 0.38039216, 1.0]],
                [[0.82745098, 0.82745098, 0.82745098, 1.0]],
                [[0.5372549, 0.81176471, 0.94117647, 1.0]],
                [[0.17254902, 0.48235294, 0.71372549, 1.0]],
            ]
            lens = [38, 3, 277, 3, 82]
            for i, col in enumerate(ax.collections):
                np.testing.assert_array_almost_equal(col.get_facecolors(), expected[i])
                assert len(col.get_paths()) == lens[i]
        else:
            unique, counts = np.unique(
                ax.collections[0].get_facecolors(), axis=0, return_counts=True
            )
            np.testing.assert_array_almost_equal(
                unique,
                np.array(
                    [
                        [0.17254902, 0.48235294, 0.71372549, 1.0],
                        [0.5372549, 0.81176471, 0.94117647, 1.0],
                        [0.82745098, 0.82745098, 0.82745098, 1.0],
                        [0.84313725, 0.09803922, 0.10980392, 1.0],
                        [0.99215686, 0.68235294, 0.38039216, 1.0],
                    ]
                ),
            )

            np.testing.assert_array_equal(counts, np.array([86, 3, 298, 38, 3]))

    @parametrize_sac
    def test_plot_scatter(self, w):
        import matplotlib

        matplotlib.use("Agg")

        lm = moran.Moran_Local(
            sac1.WHITE,
            w,
            transformation="r",
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )

        ax = lm.plot_scatter()

        # test scatter
        unique, counts = np.unique(
            ax.collections[0].get_facecolors(), axis=0, return_counts=True
        )
        np.testing.assert_array_almost_equal(
            unique,
            np.array(
                [
                    [0.17254902, 0.48235294, 0.71372549, 0.6],
                    [0.5372549, 0.81176471, 0.94117647, 0.6],
                    [0.82745098, 0.82745098, 0.82745098, 0.6],
                    [0.84313725, 0.09803922, 0.10980392, 0.6],
                    [0.99215686, 0.68235294, 0.38039216, 0.6],
                ]
            ),
        )
        np.testing.assert_array_equal(counts, np.array([73, 12, 261, 52, 5]))

        # test fitline
        l_ = ax.lines[2]
        x, y = l_.get_data()
        np.testing.assert_almost_equal(x.min(), -1.8236414387225368)
        np.testing.assert_almost_equal(x.max(), 3.893056527659032)
        np.testing.assert_almost_equal(y.min(), -0.7371749399524187)
        np.testing.assert_almost_equal(y.max(), 1.634939204358587)
        assert l_.get_color() == "k"

    @parametrize_sac
    def test_plot_scatter_args(self, w):
        import matplotlib

        matplotlib.use("Agg")

        lm = moran.Moran_Local(
            sac1.WHITE,
            w,
            transformation="r",
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )

        ax = lm.plot_scatter(
            crit_value=None,
            scatter_kwds={"s": 10},
            fitline_kwds={"linewidth": 4},
        )
        # test scatter
        np.testing.assert_array_almost_equal(
            ax.collections[0].get_facecolors(),
            np.array([[0.729412, 0.729412, 0.729412, 0.6]]),
        )
        assert ax.collections[0].get_sizes()[0] == 10

        # test fitline
        l_ = ax.lines[2]
        assert l_.get_color() == "#d6604d"
        assert l_.get_linewidth() == 4.0

    @parametrize_desmith
    def test_parallel(self, w):
        lm = moran.Moran_Local(
            self.y,
            w,
            transformation="r",
            n_jobs=-1,
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )
        np.testing.assert_allclose(lm.z_sim[0], -0.6990291160835514)
        np.testing.assert_allclose(lm.p_z_sim[0], 0.24226691753791396)

    @pytest.mark.skip("This function is being deprecated in the next release.")
    def test_by_col(self):
        import pandas as pd

        df = pd.DataFrame(self.y, columns=["z"])
        lm = moran.Moran_Local.by_col(
            df,
            ["z"],
            w=self.w,
            transformation="r",
            permutations=99,
            outvals=["z_sim", "p_z_sim"],
            keep_simulations=True,
            seed=SEED,
        )
        np.testing.assert_allclose(lm.z_z_sim[0], -0.6990291160835514)
        np.testing.assert_allclose(lm.z_p_z_sim[0], 0.24226691753791396)

    @parametrize_desmith
    def test_local_moments(self, w):
        lm = moran.Moran_Local(
            self.y,
            w,
            transformation="r",
            permutations=0,
            seed=SEED,
        )

        wikh_fast = moran._wikh_fast(lm.w.sparse)
        wikh_slow = moran._wikh_slow(lm.w.sparse)
        # wikh_fast_c = moran._wikh_fast(lm.w.sparse, sokal_correction=True)
        # wikh_slow_c = moran._wikh_slow(lm.w.sparse, sokal_correction=True)

        np.testing.assert_allclose(wikh_fast, wikh_slow, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(wikh_fast, wikh_slow, rtol=RTOL, atol=ATOL)
        EIc = np.array(
            [
                -0.00838113,
                -0.0243949,
                -0.07031778,
                -0.21520869,
                -0.16547163,
                -0.00178435,
                -0.11531888,
                -0.36138555,
                -0.05471258,
                -0.09413562,
            ]
        )
        VIc = np.array(
            [
                0.03636013,
                0.10412408,
                0.28600769,
                0.26389674,
                0.21576683,
                0.00779261,
                0.44633942,
                0.57696508,
                0.12929777,
                0.3730742,
            ]
        )

        EI = -np.ones((10,)) / 9
        VI = np.array(
            [
                0.394432,
                0.394432,
                0.394432,
                0.148043,
                0.148043,
                0.394432,
                0.394432,
                0.230173,
                0.230173,
                0.394432,
            ]
        )

        np.testing.assert_allclose(lm.EIc, EIc, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(lm.VIc, VIc, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(lm.EI, EI, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(lm.VI, VI, rtol=RTOL, atol=ATOL)

    @parametrize_sac
    def test_plot_combination(self, w):
        import matplotlib

        matplotlib.use("Agg")

        lm = moran.Moran_Local(
            sac1.WHITE,
            w,
            transformation="r",
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )
        axs = lm.plot_combination(
            sac1,
            "WHITE",
            legend_kwds=dict(loc="lower right"),
            region_column="FIPS",
            mask=["06067009504", "06067009503"],
            quadrant=1,
        )

        assert len(axs) == 3
        assert len(axs[0].patches) == 1

        if GPD_GE_120:
            assert len(axs[1].collections) == 8
            assert len(axs[2].collections) == 8
        else:
            assert len(axs[1].collections) == 4
            assert len(axs[2].collections) == 4

        axs2 = lm.plot_combination(
            sac1,
            "WHITE",
            legend_kwds=dict(loc="lower right"),
        )

        assert len(axs2) == 3
        assert len(axs2[0].patches) == 0

        if GPD_GE_120:
            assert len(axs2[1].collections) == 5
            assert len(axs2[2].collections) == 5
        else:
            assert len(axs2[1].collections) == 1
            assert len(axs2[2].collections) == 1


class TestMoranLocalBV:
    def setup_method(self):
        f = libpysal.io.open(libpysal.examples.get_path("sids2.dbf"))
        self.gdf = gpd.read_file(libpysal.examples.get_path("sids2.shp"))
        self.x = np.array(f.by_col["SIDR79"])
        self.y = np.array(f.by_col["SIDR74"])

    @parametrize_sids
    def test_defaults(self, w):
        lm = moran.Moran_Local_BV(
            self.x,
            self.y,
            w,
            keep_simulations=True,
            transformation="r",
            permutations=99,
            seed=SEED,
        )
        np.testing.assert_allclose(lm.Is[0], 1.4649221250620736)
        np.testing.assert_allclose(lm.z_sim[0], 1.330673752886702)
        np.testing.assert_allclose(lm.p_z_sim[0], 0.09164819151535242)

    @pytest.mark.skip("This function is being deprecated in the next release.")
    def test_by_col(self):
        from libpysal.io import geotable as pdio

        df = pdio.read_files(libpysal.examples.get_path("sids2.dbf"))
        moran.Moran_Local_BV.by_col(
            df,
            ["SIDR74", "SIDR79"],
            w=self.w,
            inplace=True,
            outvals=["z_sim", "p_z_sim"],
            transformation="r",
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )
        bvstats = df["SIDR79-SIDR74_moran_local_bv"].values
        bvz = df["SIDR79-SIDR74_z_sim"].values
        bvzp = df["SIDR79-SIDR74_p_z_sim"].values
        np.testing.assert_allclose(bvstats[0], 1.4649221250620736)
        np.testing.assert_allclose(bvz[0], 1.7900932313425777, 5)
        np.testing.assert_allclose(bvzp[0], 0.036719462378528744, 5)

    @parametrize_sids
    def test_labels(self, w):
        lm = moran.Moran_Local_BV(
            self.x,
            self.y,
            w,
            transformation="r",
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )
        expected_labels = np.array(
            [
                "Insignificant",
                "Insignificant",
                "Low-Low",
                "High-Low",
                "Low-High",
                "Insignificant",
                "Insignificant",
                "Insignificant",
                "Insignificant",
                "Insignificant",
            ]
        )
        assert_array_equal(lm.get_cluster_labels()[:10], expected_labels)
        assert_array_equal(
            pd.Series(lm.get_cluster_labels(0.05)).value_counts().values,
            np.array([80, 7, 6, 5, 2]),
        )

    @parametrize_sids
    def test_explore(self, w):
        lm = moran.Moran_Local_BV(
            self.x,
            self.y,
            w,
            transformation="r",
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )
        m = lm.explore(self.gdf)
        np.testing.assert_array_equal(
            m.get_bounds(),
            [
                [33.88199234008789, -84.3238525390625],
                [36.58964920043945, -75.45697784423828],
            ],
        )
        assert len(m.to_dict()["children"]) == 2

        out_str = _fetch_map_string(m)

        assert '"High-High","__folium_color":"#d7191c"' in out_str
        assert '"Low-High","__folium_color":"#89cff0"' in out_str
        assert '"Low-Low","__folium_color":"#2c7bb6"' in out_str
        assert '"High-Low","__folium_color":"#fdae61"' in out_str
        assert '"Insignificant","__folium_color":"#d3d3d3"' in out_str

        assert out_str.count("#d7191c") == 10
        assert out_str.count("#89cff0") == 5
        assert out_str.count("#2c7bb6") == 9
        assert out_str.count("#fdae61") == 8
        assert out_str.count("#d3d3d3") == 83

    @parametrize_sids
    def test_plot(self, w):
        import matplotlib

        matplotlib.use("Agg")

        lm = moran.Moran_Local_BV(
            self.x,
            self.y,
            w,
            transformation="r",
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )
        ax = lm.plot(self.gdf)

        if GPD_GE_120:
            expected = [
                [[0.84313725, 0.09803922, 0.10980392, 1.0]],
                [[0.99215686, 0.68235294, 0.38039216, 1.0]],
                [[0.82745098, 0.82745098, 0.82745098, 1.0]],
                [[0.5372549, 0.81176471, 0.94117647, 1.0]],
                [[0.17254902, 0.48235294, 0.71372549, 1.0]],
            ]
            lens = [7, 5, 80, 2, 6]
            for i, col in enumerate(ax.collections):
                np.testing.assert_array_almost_equal(col.get_facecolors(), expected[i])
                assert len(col.get_paths()) == lens[i]
        else:
            unique, counts = np.unique(
                ax.collections[0].get_facecolors(), axis=0, return_counts=True
            )
            np.testing.assert_array_almost_equal(
                unique,
                np.array(
                    [
                        [0.17254902, 0.48235294, 0.71372549, 1.0],
                        [0.5372549, 0.81176471, 0.94117647, 1.0],
                        [0.82745098, 0.82745098, 0.82745098, 1.0],
                        [0.84313725, 0.09803922, 0.10980392, 1.0],
                        [0.99215686, 0.68235294, 0.38039216, 1.0],
                    ]
                ),
            )
            np.testing.assert_array_equal(counts, np.array([6, 2, 86, 7, 7]))

    @parametrize_sids
    def test_plot_combination(self, w):
        import matplotlib

        matplotlib.use("Agg")

        lm = moran.Moran_Local_BV(
            self.x,
            self.y,
            w,
            transformation="r",
            permutations=99,
            keep_simulations=True,
            seed=SEED,
        )
        axs = lm.plot_combination(
            self.gdf,
            "SIDR79",
            legend_kwds=dict(loc="lower right"),
            quadrant=1,
        )

        assert len(axs) == 3
        assert len(axs[0].patches) == 1
        if GPD_GE_120:
            assert len(axs[1].collections) == 7
            assert len(axs[2].collections) == 7
        else:
            assert len(axs[1].collections) == 3
            assert len(axs[2].collections) == 3


class TestMoranLocalRate:
    def setup_method(self):
        f = libpysal.io.open(libpysal.examples.get_path("sids2.dbf"))
        self.e = np.array(f.by_col["SID79"])
        self.b = np.array(f.by_col["BIR79"])

    @parametrize_sids
    def test_moran_rate(self, w):
        lm = moran.Moran_Local_Rate(
            self.e, self.b, w, transformation="r", permutations=99, seed=SEED
        )
        np.testing.assert_allclose(lm.z_sim[0], 0.02702781851384379, 7)
        np.testing.assert_allclose(lm.p_z_sim[0], 0.4892187730835096)

    @pytest.mark.skip("This function is being deprecated in the next release.")
    def test_by_col(self):
        from libpysal.io import geotable as pdio

        df = pdio.read_files(libpysal.examples.get_path("sids2.dbf"))
        lm = moran.Moran_Local_Rate.by_col(
            df,
            ["SID79"],
            ["BIR79"],
            w=self.w,
            outvals=["p_z_sim", "z_sim"],
            transformation="r",
            permutations=99,
            seed=SEED,
        )
        np.testing.assert_allclose(lm["SID79-BIR79_z_sim"][0], 0.02702781851384379, 7)
        np.testing.assert_allclose(lm["SID79-BIR79_p_z_sim"][0], 0.4892187730835096)


def _fetch_map_string(m):
    out = m._parent.render()
    out_str = "".join(out.split())
    return out_str



import numpy
from scipy import sparse
from sklearn import preprocessing, utils
from sklearn.base import BaseEstimator

from .crand import _prepare_bivariate
from .crand import njit as _njit


class Spatial_Pearson(BaseEstimator):  # noqa: N801
    """Global Spatial Pearson Statistic"""

    def __init__(self, connectivity=None, permutations=999):
        """
        Initialize a spatial pearson estimator

        Parameters
        ----------
        connectivity:   scipy.sparse matrix object
                        the connectivity structure describing the relationships
                        between observed units. Will be row-standardized.
        permutations:   int
                        the number of permutations to conduct for inference.
                        if < 1, no permutational inference will be conducted.

        Attributes
        ----------
        association_: numpy.ndarray (2,2)
                      array containg the estimated Lee spatial pearson correlation
                      coefficients, where element [0,1] is the spatial correlation
                      coefficient, and elements [0,0] and [1,1] are the "spatial
                      smoothing factor"
        reference_distribution_: numpy.ndarray (n_permutations, 2,2)
                      distribution of correlation matrices for randomly-shuffled
                      maps.
        significance_: numpy.ndarray (2,2)
                       permutation-based p-values for the fraction of times the
                       observed correlation was more extreme than the simulated
                       correlations.
        """
        self.connectivity = connectivity
        self.permutations = permutations

    def fit(self, x, y):
        """
        bivariate spatial pearson's R based on Eq. 18 of :cite:`Lee2001`.

        L = \\dfrac{Z^T (V^TV) Z}{1^T (V^TV) 1}

        Parameters
        ----------
        x       :   numpy.ndarray
                    array containing continuous data
        y       :   numpy.ndarray
                    array containing continuous data

        Returns
        -------
        the fitted estimator.

        Notes
        -----
        Technical details and derivations can be found in :cite:`Lee2001`.

        """
        x = utils.check_array(x)
        y = utils.check_array(y)
        Z = numpy.column_stack(
            (
                preprocessing.StandardScaler().fit_transform(x),
                preprocessing.StandardScaler().fit_transform(y),
            )
        )
        if self.connectivity is None:
            self.connectivity = sparse.eye(Z.shape[0])
        self.association_ = self._statistic(Z, self.connectivity)

        if self.permutations is None or self.permutations < 1:
            return self

        if self.permutations:
            simulations = [
                self._statistic(numpy.random.permutation(Z), self.connectivity)
                for _ in range(self.permutations)
            ]
            self.reference_distribution_ = simulations = numpy.array(simulations)
            above = simulations >= self.association_
            larger = above.sum(axis=0)
            extreme = numpy.minimum(self.permutations - larger, larger)
            self.significance_ = (extreme + 1.0) / (self.permutations + 1.0)
        return self

    @staticmethod
    def _statistic(Z, W):
        ctc = W.T @ W
        ones = numpy.ones(ctc.shape[0])
        return (Z.T @ ctc @ Z) / (ones.T @ ctc @ ones)


class Spatial_Pearson_Local(BaseEstimator):  # noqa: N801
    """Local Spatial Pearson Statistic"""

    def __init__(self, connectivity=None, permutations=999):
        """
        Initialize a spatial local pearson estimator

        Parameters
        ----------
        connectivity:   scipy.sparse matrix object
                        the connectivity structure describing the relationships
                        between observed units. Will be row-standardized.
        permutations:   int
                        the number of permutations to conduct for inference.
                        if < 1, no permutational inference will be conducted.
        significance_: numpy.ndarray (2,2)
                       permutation-based p-values for the fraction of times the
                       observed correlation was more extreme than the simulated
                       correlations.
        Attributes
        ----------
        associations_: numpy.ndarray (n_samples,)
                      array containg the estimated Lee spatial pearson correlation
                      coefficients, where element [0,1] is the spatial correlation
                      coefficient, and elements [0,0] and [1,1] are the "spatial
                      smoothing factor"
        reference_distribution_: numpy.ndarray (n_permutations, n_samples)
                      distribution of correlation matrices for randomly-shuffled
                      maps.
        significance_: numpy.ndarray (n_samples,)
                       permutation-based p-values for the fraction of times the
                       observed correlation was more extreme than the simulated
                       correlations.

        Notes
        -----
        Technical details and derivations can be found in :cite:`Lee2001`.

        """
        self.connectivity = connectivity
        self.permutations = permutations

    def fit(self, x, y):
        """
        Bivariate local pearson's R based on Eq. 22 in Lee (2001), using
        site-wise conditional randomization from Moran_Local_BV.

        .. math::

            L_i = \\dfrac{
                         n \\cdot
                           \\Big[\big(\\sum_i w_{ij}(x_j - \bar{x})\big)
                                 \big(\\sum_i w_{ij}(y_j - \bar{y})\big) \\Big]
                         }
                        {
                         \\sqrt{\\sum_i (x_i - \bar{x})^2}
                         \\sqrt{\\sum_i (y_i - \bar{y})^2}}
                = \\dfrac{
                         n \\cdot
                           (\tilde{x}_j - \bar{x})
                           (\tilde{y}_j - \bar{y})
                         }
                        {
                         \\sqrt{\\sum_i (x_i - \bar{x})^2}
                         \\sqrt{\\sum_i (y_i - \bar{y})^2}}

        Lee, Sang Il. (2001), "Developing a bivariate spatial
        association measure: An integration of Pearson's r and
        Moran's I." Journal of Geographical Systems, 3(4):369-385.

        Parameters
        ----------
        x       :   numpy.ndarray
                    array containing continuous data
        y       :   numpy.ndarray
                    array containing continuous data

        Returns
        -------
        the fitted estimator.
        """
        x = utils.check_array(x)
        x = preprocessing.StandardScaler().fit_transform(x)

        y = utils.check_array(y)
        y = preprocessing.StandardScaler().fit_transform(y)

        Z = numpy.column_stack((x, y))

        standard_connectivity = sparse.csc_matrix(
            self.connectivity / self.connectivity.sum(axis=1).reshape(-1, 1)
        )

        n, _ = x.shape

        self.associations_ = self._statistic(Z, standard_connectivity)

        if self.permutations:
            self.reference_distribution_ = numpy.empty((n, self.permutations))
            max_neighbors = (standard_connectivity != 0).sum(axis=1).max()
            random_ids = numpy.array(
                [
                    numpy.random.permutation(n - 1)[0 : max_neighbors + 1]
                    for i in range(self.permutations)
                ]
            )
            ids = numpy.arange(n)

            for i in range(n):
                row = standard_connectivity[i]
                weight = numpy.asarray(row[row.nonzero()]).reshape(-1, 1)
                cardinality = row.nonzero()[0].shape[0]

                ids_not_i = ids[ids != i]
                numpy.random.shuffle(ids_not_i)
                randomizer = random_ids[:, 0:cardinality]
                random_neighbors = ids_not_i[randomizer]

                random_neighbor_x = x[random_neighbors]
                random_neighbor_y = y[random_neighbors]

                self.reference_distribution_[i] = (
                    (weight * random_neighbor_y).sum(axis=1) - y.mean()
                ).squeeze()
                self.reference_distribution_[i] *= (
                    (weight * random_neighbor_x).sum(axis=1) - x.mean()
                ).squeeze()
            above = self.reference_distribution_ >= self.associations_.reshape(-1, 1)
            larger = above.sum(axis=1)
            extreme = numpy.minimum(larger, self.permutations - larger)
            self.significance_ = (extreme + 1.0) / (self.permutations + 1.0)
            self.reference_distribution_ = self.reference_distribution_.T
        else:
            self.reference_distribution_ = None
        return self

    @staticmethod
    def _statistic(Z, W):
        return (Z[:, 1] @ W.T) * (W @ Z[:, 0])


# --------------------------------------------------------------
# Conditional Randomization Function Implementations
# --------------------------------------------------------------


@_njit(fastmath=True)
def _local_spatial_pearson_crand(i, z, permuted_ids, weights_i, scaling):
    zxi, zxrand, zyi, zyrand = _prepare_bivariate(i, z, permuted_ids, weights_i)
    return (zyrand @ weights_i) * (zxrand @ weights_i) * scaling


if __name__ == "__main__":
    import geopandas
    import libpysal

    df = geopandas.read_file(libpysal.examples.get_path("columbus.shp"))
    x = df[["HOVAL"]].values
    y = df[["CRIME"]].values
    zx = preprocessing.StandardScaler().fit_transform(x)
    zy = preprocessing.StandardScaler().fit_transform(y)
    w = libpysal.weights.Queen.from_dataframe(df)
    w.transform = "r"
    numpy.random.seed(2478879)
    testglobal = Spatial_Pearson(connectivity=w.sparse).fit(x, y)
    numpy.random.seed(2478879)
    testlocal = Spatial_Pearson_Local(connectivity=w.sparse).fit(x, y)



import numpy
import pandas
from libpysal import weights
from scipy.spatial import distance
from scipy.stats import mode as most_common_value
from sklearn.utils import check_array

try:
    from tqdm import tqdm

    HAS_TQDM = True
except ImportError:
    HAS_TQDM = False


def _passthrough(sequence):
    return sequence


def _resolve_metric(X, coordinates, metric):
    """
    Provide a distance function that you can use
    to find the distance betwen arbitrary points.
    """
    if callable(metric):
        distance_func = metric
    elif metric.lower() == "haversine":
        try:
            from numba import autojit
        except ImportError:

            def autojit(func):
                return func

        @autojit
        def harcdist(p1, p2):
            """Compute the kernel of haversine"""
            x = numpy.sin(p2[1] - p1[1] / 2) ** 2
            cosp1, cosp2 = numpy.cos(p1[1]), numpy.cos(p2[1])
            y = cosp2 * cosp1 * numpy.sin((p2[0] - p1[0]) / 2) ** 2
            return 2 * numpy.arcsin(numpy.sqrt(x + y))

        distance_func = harcdist
    elif metric.lower() == "precomputed":
        # so, in this case, coordinates is actually distance matrix of some kind
        # and is assumed aligned to X, such that the distance from X[a] to X[b] is
        # coordinates[a,b], confusingly... So, we'll re-write them as "distances"
        distances = check_array(coordinates, accept_sparse=True)
        n, k = distances.shape
        assert k == n, (
            "With metric='precomputed', coordinates must be an (n,n)"
            " matrix representing distances between coordinates."
        )

        def lookup_distance(a, b):
            """Find location of points a,b in X and return precomputed distances"""
            (aloc,) = (a == X).all(axis=1).nonzero()
            (bloc,) = (b == X).all(axis=1).nonzero()
            if (len(aloc) > 1) or (len(bloc) > 1):
                raise NotImplementedError(
                    "Precomputed distances cannot disambiguate coincident points."
                    " Add a slight bit of noise to the input to force them"
                    " into non-coincidence and re-compute the distance matrix."
                )
            elif (len(aloc) == 0) or (len(bloc) == 0):
                raise NotImplementedError(
                    "Precomputed distances cannot compute distances to new points."
                )
            return distances[aloc, bloc]

        distance_func = lookup_distance
    else:
        try:
            distance_func = getattr(distance, metric)
        except AttributeError:
            raise KeyError(
                f"Metric '{metric}' not understood. Choose "
                "something available in scipy.spatial.distance."
            ) from None
    return distance_func


def isolation(
    X,
    coordinates,
    metric="euclidean",
    middle="mean",
    return_all=False,
    progressbar=False,
):
    """
    Compute the isolation of each value of X by constructing the distance
    to the nearest higher value in the data.

    Parameters
    ----------
    X : numpy.ndarray
        (N, p) array of data to use as input. If p > 1, the "elevation" is computed
        using the topo.to_elevation function.
    coordinates : numpy.ndarray
        (N,k) array of locations for X to compute distances. If
        metric='precomputed', this should contain the distances from
        each point to every other point, and k == N.
    metric : string or callable (default: 'euclidean')
        name of distance metric in scipy.spatial.distance, or function, that can be
        used to compute distances between locations. If 'precomputed', ad-hoc function
        will be defined to look up distances between points instead.
    middle : string or callable (default: 'mean')
        method to define the elevation of points. See to_elevation for more details.
    return_all : bool (default: False)
        if False, only return the isolation (distance to nearest higher value).
    progressbar: bool (default: False)
        if True, show a progressbar for the computation.
    Returns
    -------
    either (N,) array of isolation values, or a pandas dataframe containing the full
    tree of precedence for the isolation tree.
    """
    X = check_array(X, ensure_2d=False)
    X = to_elevation(X, middle=middle).squeeze()
    try:
        from rtree.index import Index as SpatialIndex
    except ImportError:
        raise ImportError(
            "rtree library must be installed to use the prominence measure"
        ) from None
    distance_func = _resolve_metric(X, coordinates, metric)
    sort_order = numpy.argsort(-X)
    tree = SpatialIndex()
    ix = sort_order[0]
    tree.insert(0, tuple(coordinates[ix]), obj=X[ix])
    precedence_tree = [[ix, numpy.nan, 0, numpy.nan, numpy.nan, numpy.nan]]

    if progressbar and HAS_TQDM:
        pbar = tqdm
    elif progressbar and (not HAS_TQDM):
        raise ImportError("The `tqdm` module is required for progressbars.")
    else:
        pbar = _passthrough

    for iter_ix, ix in pbar(enumerate(sort_order[1:])):
        rank = iter_ix + 1
        value = X[ix]
        location = coordinates[ix,]
        (match,) = tree.nearest(tuple(location), objects=True)
        higher_rank = match.id
        higher_value = match.object
        higher_location = match.bbox[:2]
        higher_ix = sort_order[higher_rank]
        distance = distance_func(location, higher_location)
        gap = higher_value - value
        precedence_tree.append([ix, higher_ix, rank, higher_rank, distance, gap])
        tree.insert(rank, tuple(location), obj=value)
    # return precedence_tree
    precedence_tree = numpy.asarray(precedence_tree)
    # print(precedence_tree.shape)
    out = numpy.empty_like(precedence_tree)
    out[sort_order] = precedence_tree
    result = pandas.DataFrame(
        out,
        columns=["index", "parent_index", "rank", "parent_rank", "isolation", "gap"],
    ).sort_values(["index", "parent_index"])
    if return_all:
        return result
    else:
        return result.isolation.values


def prominence(
    X,
    connectivity,
    return_all=False,
    gdf=None,
    verbose=False,
    middle="mean",
    progressbar=False,
):
    """
    Return the prominence of peaks in input, given a connectivity matrix.

    Parameters
    ----------
    X : numpy.ndarray
        an array of shape N,p containing data to use for computing prominence. When
        p > 1, X will be converted to an "elevation" using to_elevation.
    connectivity : scipy.sparse matrix
        a sparse matrix encoding the connectivity graph pertaining to rows of X. If
        coordinates are provided, they must be (N,2), and the delaunay triangulation
        will be computed.
    return_class : bool (default: False)
        whether or not to return additional information about the result, such as
        the set of dominating peaks or the set of classifications for each observation.
    verbose : bool (default: None)
        whether or not to print extra information about the progress of the algorithm.
    middle : string or callable (default: "mean")
        how to compute the center of mass from X, when the dimension of X > 2.

    Returns
    -------
    the prominence of each observation in X, possibly along with the
    set of saddle points, peaks, and/or dominating peak tree.

    Notes
    -----
    An observation has 0 prominence when it is a saddle point.
    An observation has positive prominence when it is a peak, and this
    is computed as the elevation of the peak minus the elevation of the saddle point.

    Observations have "NA" prominence when they are neither a saddle point nor a peak.

    """
    X = check_array(X, ensure_2d=False).squeeze()
    X = to_elevation(X, middle=middle).squeeze()
    (n,) = X.shape

    if not isinstance(verbose, bool | int):
        gdf = verbose
        verbose = True
    else:
        gdf = None

    connectivity = _check_connectivity(connectivity)

    # sort the variable in descending order
    sort_order = numpy.argsort(-X)

    peaks = []
    assessed_peaks = set()
    prominence = numpy.empty_like(X) * numpy.nan
    dominating_peak = numpy.ones_like(X) * -1
    predecessors = numpy.ones_like(X) * -1
    keycols = numpy.ones_like(X) * -1
    ids = numpy.arange(n)
    classifications = [None] * n
    key_cols = dict()

    if progressbar and HAS_TQDM:
        pbar = tqdm
    elif progressbar and (not HAS_TQDM):
        raise ImportError("The `tqdm` module is required for progressbars.")
    else:
        pbar = _passthrough

    for rank, value in pbar(enumerate(X[sort_order])):
        # This is needed to break ties in the same way that argsort does. A more
        # natural way to do this is to use X >= value, but if value is tied, then
        # that would generate a mask where too many elements are selected!
        # e.g. mask.sum() > rank
        mask = numpy.isin(numpy.arange(n), sort_order[: rank + 1])
        (full_indices,) = mask.nonzero()
        this_full_ix = (ids[sort_order])[rank]
        msg = f"assessing {this_full_ix} (rank: {rank}, value: {value})"

        # use the dominating_peak vector. A new obs either has:
        #   Neighbors whose dominating_peak...
        #       1. ... are all -1 (new peak)
        #       2. ... are all -1 or an integer (slope of current peak)
        #       3. ... include at least two integers and any -1 (key col)
        _, neighbs = connectivity[this_full_ix].toarray().nonzero()
        this_preds = predecessors[neighbs]

        # need to keep ordering in this sublist to preserve hierarchy
        this_unique_preds = [p for p in peaks if ((p in this_preds) & (p >= 0))]
        joins_new_subgraph = not set(this_unique_preds).issubset(assessed_peaks)
        if tuple(this_unique_preds) in key_cols:
            classification = "slope"
        elif len(this_unique_preds) == 0:
            classification = "peak"
        elif (len(this_unique_preds) >= 2) & joins_new_subgraph:
            classification = "keycol"
        else:
            classification = "slope"

        classifications[this_full_ix] = classification

        if (
            classification == "keycol"
        ):  # this_ix merges two or more subgraphs, so is a key_col
            # find the peaks it joins
            now_joined_peaks = this_unique_preds
            # add them as keys for the key_col lut
            key_cols.update({tuple(now_joined_peaks): this_full_ix})
            msg += f"\n{this_full_ix} is a key col between {now_joined_peaks}!"
            dominating_peak[this_full_ix] = now_joined_peaks[
                -1
            ]  # lowest now-joined peak
            predecessors[this_full_ix] = now_joined_peaks[-1]
            prominence[this_full_ix] = 0
            # given we now know the key col, get the prominence for
            # unassayed peaks in the subgraph
            for peak_ix in now_joined_peaks:
                if peak_ix in assessed_peaks:
                    continue
                # prominence is peak - key col
                keycols[peak_ix] = this_full_ix
                prominence[peak_ix] -= value
                assessed_peaks.update((peak_ix,))
        elif classification == "peak":  # this_ix is a new peak since it's disconnected
            msg += f"\n{this_full_ix} is a peak!"
            # its parent is the last visited peak (for precedence purposes)
            try:
                previous_peak = peaks[-1]
            except IndexError:
                previous_peak = this_full_ix
            if this_full_ix not in peaks:
                peaks.append(this_full_ix)
            dominating_peak[this_full_ix] = previous_peak
            predecessors[this_full_ix] = this_full_ix
            # we initialize prominence here, rather than compute it solely in
            # the `key_col` branch because a graph `island` disconnected observation
            # should have prominence "value - 0", since it has no key cols
            prominence[this_full_ix] = X[this_full_ix]
        else:  # this_ix is connected to an existing peak, but doesn't bridge peaks.
            msg += f"\n{this_full_ix} is a slope!"
            # get all the peaks that are linked to this slope
            this_peak = this_unique_preds
            if len(this_peak) == 1:  # if there's only one peak the slope is related to
                # then use it
                best_peak = this_peak[0]
            else:  # otherwise, if there are multiple peaks
                # pick the one that most of its neighbors are assigned to
                best_peak = most_common_value(
                    this_unique_preds, keepdims=False
                ).mode.item()
            all_on_slope = numpy.arange(n)[dominating_peak == best_peak]
            msg += f"\n{all_on_slope} are on the slopes of {best_peak}."
            dominating_peak[this_full_ix] = best_peak
            predecessors[this_full_ix] = best_peak

        if verbose:
            print(
                "--------------------------------------------\n"
                f"at the {rank} iteration:\n{msg}\n\tpeaks\t{peaks}\n"
                f"\tprominence\t{prominence}\n\tkey_cols\t{key_cols}\n"
            )
        if gdf is not None:
            peakframe = gdf.iloc[peaks]
            keycolframe = gdf.iloc[list(key_cols.values())]
            _isin_peak = gdf.index.isin(peakframe.index)
            _isin_keycol = gdf.index.isin(keycolframe.index)
            slopeframe = gdf[~(_isin_peak | _isin_keycol) & mask]
            rest = gdf[~mask]
            this_geom = gdf.iloc[[this_full_ix]]
            ax = rest.plot(edgecolor="k", lw=0.1, fc="lightblue")
            ax = slopeframe.plot(ec="k", lw=0.1, fc="linen", ax=ax)
            ax = keycolframe.plot(ec="k", lw=0.1, fc="red", ax=ax)
            ax = peakframe.plot(ec="k", lw=0.1, fc="yellow", ax=ax)
            ax = this_geom.centroid.plot(ax=ax, color="orange", marker="*")
            plt.show()
            command = input()
            if command.strip().lower() == "stop":
                break
    result = pandas.DataFrame.from_dict(
        dict(
            index=ids,
            prominence=prominence,
            classification=classifications,
            predecessor=predecessors,
            keycol=keycols,
            dominating_peak=dominating_peak,
        )
    )
    if not return_all:
        return result.prominence.values
    return result


def to_elevation(X, middle="mean", metric="euclidean"):
    """
    Compute the "elevation" of coordinates in p-dimensional space.

    For 1 dimensional X, this simply sets the zero point at the minimum value
    for the data. As an analogue to physical elevation, this means that the
    lowest value in 1-dimensional X is considered "sea level."

    For X in higher dimension, we treat X as defining a location on a (hyper)sphere.
    The "elevation," then, is the distance from the center of mass.
    So, this computes the distance of each point to the overall the center of mass
    and uses this as the "elevation," setting sea level (zero) to the lowest elevation.

    Parameters
    ----------
    X : numpy.ndarray
        Array of values for which to compute elevation.
    middle : callable or string
        name of function in numpy (or function itself)
        used to compute the center point of X
    metric : string
        metric to use in `scipy.spatial.distance.cdist` to
        compute the distance from the center of mass to the point.

    Returns
    --------
    (N,1)-shaped numpy array containing the "elevation" of
    each point relative to sea level (zero).

    """
    if X.ndim == 1:
        return X - X.min()
    else:
        if callable(middle):
            middle_point = middle(X, axis=0)
        else:
            try:
                middle = getattr(numpy, middle)
                return to_elevation(X, middle=middle)
            except AttributeError:
                raise KeyError(
                    f"numpy has no '{middle}' function to "
                    "compute the middle of a point cloud."
                ) from None
        distance_from_center = distance.cdist(
            X, middle_point.reshape(1, -1), metric=metric
        )
        return to_elevation(distance_from_center.squeeze())


def _check_connectivity(connectivity_or_coordinates):
    """
    Check that connectivity provided is either:
    1. a sparse graph from scipy.sparse
    2. a weights object from libpysal that we need to cast to a scipy.sparse matrix
    3. a set of coordinates that we need to build the delaunay triangulation
       and then return the graph.
    """
    from scipy.sparse import issparse

    if issparse(connectivity_or_coordinates):
        shape = connectivity_or_coordinates.shape
        assert shape[0] == shape[1], (
            f"Connectivity matrix must be square, but is {shape}."
        )
        return connectivity_or_coordinates
    if issubclass(type(connectivity_or_coordinates), weights.W):
        return connectivity_or_coordinates.sparse
    else:
        from libpysal.weights import Voronoi

        return _check_connectivity(Voronoi(connectivity_or_coordinates))


if __name__ == "__main__":
    import geopandas
    import matplotlib.pyplot as plt
    import pandas  # noqa: F811
    from libpysal import examples, weights  # noqa: F811
    from matplotlib import cm

    current_cmap = cm.get_cmap()
    current_cmap.set_bad(color="lightgrey")

    data = (
        geopandas.read_file(examples.get_path("NAT.shp"))
        .query('STATE_NAME == "Illinois"')
        .reset_index()
    )
    coordinates = numpy.column_stack((data.centroid.x, data.centroid.y))
    gini = data[["GI89"]].values.flatten()
    contig_graph = weights.Rook.from_dataframe(data)
    iso = isolation(gini, coordinates, return_all=True)

    f, ax = plt.subplots(1, 3)
    for i in range(1, 3):
        data.plot(color="lightgrey", ax=ax[i])
    data.plot(iso.gap, ax=ax[1], cmap=current_cmap)
    data.plot(iso.distance, ax=ax[2], cmap=current_cmap)
    data.plot("GI89", ax=ax[0])
    ax[0].set_title("Variable")
    ax[1].set_title("Releif")
    ax[2].set_title("Isolation")
    plt.show()

    prom = prominence(gini, contig_graph)
    ax = data.plot(prom)
    data.plot(color="lightgrey", ax=ax, zorder=-1)
    plt.show()



import numpy as np
import pandas as pd
from libpysal import weights
from sklearn.base import BaseEstimator

from esda.crand import _prepare_bivariate, _prepare_univariate
from esda.crand import crand as _crand_plus
from esda.crand import njit as _njit

PERMUTATIONS = 999


class Join_Counts_Local_BV(BaseEstimator):  # noqa: N801
    """Univariate Local Join Count Statistic"""

    def __init__(
        self,
        connectivity=None,
        permutations=PERMUTATIONS,
        n_jobs=1,
        keep_simulations=True,
        seed=None,
        island_weight=0,
        drop_islands=True,
    ):
        """
        Initialize a Local_Join_Counts_BV estimator

        Parameters
        ----------
        connectivity : W | Graph
            spatial weights instance as W or Graph aligned with y
        permutations : int
            number of random permutations for calculation of pseudo p_values
        n_jobs : int
            Number of cores to be used in the conditional randomisation.
            If -1, all available cores are used.
        keep_simulations : bool (default True)
            If True, the entire matrix of replications under the null
            is stored in memory and accessible; otherwise, replications
            are not saved
        seed : None/int
            Seed to ensure reproducibility of conditional randomizations.
            Must be set here, and not outside of the function, since numba
            does not correctly interpret external seeds
            nor numpy.random.RandomState instances.
        island_weight:
            value to use as a weight for the "fake" neighbor for every island.
            If numpy.nan, will propagate to the final local statistic depending
            on the `stat_func`. If 0, then the lag is always zero for islands.
        drop_islands : bool (default True)
            Whether or not to preserve islands as entries in the adjacency
            list. By default, observations with no neighbors do not appear
            in the adjacency list. If islands are kept, they are coded as
            self-neighbors with zero weight. See ``libpysal.weights.to_adjlist()``.
        """

        self.connectivity = connectivity
        self.permutations = permutations
        self.n_jobs = n_jobs
        self.keep_simulations = keep_simulations
        self.seed = seed
        self.island_weight = island_weight
        self.drop_islands = drop_islands

    def fit(self, x, z, case="CLC", n_jobs=1, permutations=999):
        """
        Parameters
        ----------
        x                : numpy.ndarray
                           array containing binary (0/1) data
        z                : numpy.ndarray
                           array containing binary (0/1) data
        Returns
        -------
        the fitted estimator.

        Notes
        -----
        Technical details and derivations can be found in :cite:`AnselinLi2019`.

        Examples
        --------
        >>> import libpysal
        >>> w = libpysal.weights.lat2W(4, 4)
        >>> x = np.ones(16)
        >>> x[0:8] = 0
        >>> z = [0,1,0,1,1,1,1,1,0,0,1,1,0,0,1,1]
        >>> LJC_BV_C1 = Local_Join_Counts_BV(connectivity=w).fit(x, z, case="BJC")
        >>> LJC_BV_C2 = Local_Join_Counts_BV(connectivity=w).fit(x, z, case="CLC")
        >>> LJC_BV_C1.LJC
        >>> LJC_BV_C1.p_sim
        >>> LJC_BV_C2.LJC
        >>> LJC_BV_C2.p_sim

        Commpop data replicating GeoDa tutorial (Case 1)
        >>> import libpysal
        >>> import geopandas as gpd
        >>> commpop = gpd.read_file(
        ...     "https://github.com/jeffcsauer/GSOC2020/raw/master/"
        ...     "validation/data/commpop.gpkg"
        ... )
        >>> w = libpysal.weights.Queen.from_dataframe(commpop)
        >>> LJC_BV_Case1 = Local_Join_Counts_BV(
        ...     connectivity=w
        ... ).fit(commpop['popneg'], commpop['popplus'], case='BJC')
        >>> LJC_BV_Case1.LJC
        >>> LJC_BV_Case1.p_sim

        Guerry data replicating GeoDa tutorial (Case 2)
        >>> import libpysal
        >>> import geopandas as gpd
        >>> guerry = libpysal.examples.load_example('Guerry')
        >>> guerry_ds = gpd.read_file(guerry.get_path('Guerry.shp'))
        >>> guerry_ds['infq5'] = 0
        >>> guerry_ds['donq5'] = 0
        >>> guerry_ds.loc[(guerry_ds['Infants'] > 23574), 'infq5'] = 1
        >>> guerry_ds.loc[(guerry_ds['Donatns'] > 10973), 'donq5'] = 1
        >>> w = libpysal.weights.Queen.from_dataframe(guerry_ds)
        >>> LJC_BV_Case2 = Local_Join_Counts_BV(
        ...     connectivity=w
        ... ).fit(guerry_ds['infq5'], guerry_ds['donq5'], case='CLC')
        >>> LJC_BV_Case2.LJC
        >>> LJC_BV_Case2.p_sim
        """
        # Need to ensure that the np.array() are of dtype='float' for numba
        x = np.array(x, dtype="float")
        z = np.array(z, dtype="float")

        w = self.connectivity
        # Fill the diagonal with 0s
        if isinstance(w, weights.W):
            w = weights.util.fill_diagonal(w, val=0)
            w.transform = "b"
        else:
            w = w.assign_self_weight(0).eliminate_zeros().transform("b")

        self.x = x
        self.z = z
        self.n = len(x)
        self.w = w
        self.case = case

        n_jobs = self.n_jobs

        self.LJC = self._statistic(x, z, w, case, self.drop_islands)

        if permutations:
            if case == "BJC":
                self.p_sim, self.rjoins = _crand_plus(
                    z=np.column_stack((x, z)),
                    w=self.w,
                    observed=self.LJC,
                    permutations=permutations,
                    keep=True,
                    n_jobs=n_jobs,
                    stat_func=_ljc_bv_case1,
                    island_weight=self.island_weight,
                )
                # Set p-values for those with LJC of 0 to NaN
                self.p_sim[self.LJC == 0] = "NaN"
            elif case == "CLC":
                self.p_sim, self.rjoins = _crand_plus(
                    z=np.column_stack((x, z)),
                    w=self.w,
                    observed=self.LJC,
                    permutations=permutations,
                    keep=True,
                    n_jobs=n_jobs,
                    stat_func=_ljc_bv_case2,
                    island_weight=self.island_weight,
                )
                # Set p-values for those with LJC of 0 to NaN
                self.p_sim[self.LJC == 0] = "NaN"
            else:
                raise NotImplementedError(
                    f"The requested LJC method ({case}) is not currently supported!"
                )

        return self

    @staticmethod
    def _statistic(x, z, w, case, drop_islands):
        # Create adjacency list. Note that remove_symmetric=False - this is
        # different from the esda.Join_Counts() function.
        if isinstance(w, weights.W):
            adj_list = w.to_adjlist(remove_symmetric=False, drop_islands=drop_islands)
            # First, set up a series that maps the values to the weights table
            zseries_x = pd.Series(x, index=w.id_order)
            zseries_z = pd.Series(z, index=w.id_order)
        else:
            if drop_islands:
                adj_list = w.adjacency.drop(w.isolates).reset_index()
            else:
                adj_list = w.adjacency.reset_index()
            # First, set up a series that maps the values to the weights table
            zseries_x = pd.Series(x, index=w.unique_ids)
            zseries_z = pd.Series(z, index=w.unique_ids)

        # Map the values to the focal (i) values
        focal_x = zseries_x.loc[adj_list.focal].values
        focal_z = zseries_z.loc[adj_list.focal].values

        # Map the values to the neighbor (j) values
        neighbor_x = zseries_x.loc[adj_list.neighbor].values
        neighbor_z = zseries_z.loc[adj_list.neighbor].values

        if case == "BJC":
            BJC = (
                (focal_x == 1) & (focal_z == 0) & (neighbor_x == 0) & (neighbor_z == 1)
            )
            adj_list_BJC = pd.DataFrame(
                adj_list.focal.values, BJC.astype("uint8")
            ).reset_index()
            adj_list_BJC.columns = ["BJC", "ID"]
            adj_list_BJC = adj_list_BJC.groupby(by="ID").sum()
            return np.array(adj_list_BJC.BJC.values, dtype="float")
        elif case == "CLC":
            CLC = (
                (focal_x == 1) & (focal_z == 1) & (neighbor_x == 1) & (neighbor_z == 1)
            )
            adj_list_CLC = pd.DataFrame(
                adj_list.focal.values, CLC.astype("uint8")
            ).reset_index()
            adj_list_CLC.columns = ["CLC", "ID"]
            adj_list_CLC = adj_list_CLC.groupby(by="ID").sum()
            return np.array(adj_list_CLC.CLC.values, dtype="float")
        else:
            raise NotImplementedError(
                f"The requested LJC method ({case}) is not currently supported!"
            )


# --------------------------------------------------------------
# Conditional Randomization Function Implementations
# --------------------------------------------------------------

# Note: scaling not used


@_njit(fastmath=True)
def _ljc_bv_case1(i, z, permuted_ids, weights_i, scaling):  # noqa: ARG001
    zx = z[:, 0]
    zy = z[:, 1]
    other_weights = weights_i[1:]
    zyi, zyrand = _prepare_univariate(i, zy, permuted_ids, other_weights)
    return zx[i] * (zyrand @ other_weights)


@_njit(fastmath=True)
def _ljc_bv_case2(i, z, permuted_ids, weights_i, scaling):  # noqa: ARG001
    zy = z[:, 1]
    other_weights = weights_i[1:]
    zxi, zxrand, zyi, zyrand = _prepare_bivariate(i, z, permuted_ids, other_weights)
    zf = zxrand * zyrand
    return zy[i] * (zf @ other_weights)



import numpy as np
import pandas
import pytest
import sklearn

from .. import adbscan


class TestADBSCAN:
    def setup_method(self):
        np.random.seed(10)
        self.db = pandas.DataFrame(
            {"x": np.random.random(25), "y": np.random.random(25)}
        )
        self.lbls = np.array(
            [
                "-1",
                "-1",
                "-1",
                "0",
                "-1",
                "-1",
                "-1",
                "0",
                "-1",
                "-1",
                "-1",
                "-1",
                "-1",
                "-1",
                "0",
                "0",
                "0",
                "-1",
                "0",
                "-1",
                "0",
                "-1",
                "-1",
                "-1",
                "-1",
            ],
            dtype=object,
        )
        self.pcts = np.array(
            [
                0.7,
                0.5,
                0.7,
                1.0,
                0.7,
                0.7,
                0.5,
                1.0,
                0.7,
                0.7,
                0.6,
                0.6,
                0.6,
                0.7,
                1.0,
                0.9,
                1.0,
                0.7,
                1.0,
                0.7,
                0.9,
                0.7,
                0.8,
                0.6,
                0.7,
            ]
        )

    @pytest.mark.skipif(sklearn.__version__ == "1.3.0", reason="sklearn regression")
    def test_adbscan(self):
        # ------------------------#
        #           # Single Core #
        # ------------------------#
        np.random.seed(10)
        ads = adbscan.ADBSCAN(0.03, 3, reps=10, keep_solus=True)
        _ = ads.fit(self.db, xy=["x", "y"])
        # Params
        assert pytest.approx(ads.eps) == 0.03
        assert ads.min_samples == 3
        assert ads.algorithm == "auto"
        assert ads.n_jobs == 1
        assert ads.pct_exact == 0.1
        assert ads.reps == 10
        assert ads.keep_solus is True
        assert ads.pct_thr == 0.9
        # Labels
        np.testing.assert_equal(ads.labels_, self.lbls)
        # Votes
        votes = pandas.DataFrame({"lbls": self.lbls, "pct": self.pcts})
        np.testing.assert_equal(ads.votes["lbls"].values, self.lbls)
        np.testing.assert_almost_equal(ads.votes["pct"].values, votes["pct"].values)
        # Solus
        np.testing.assert_equal(ads.solus.astype(int).sum().sum(), 133)
        rep_sum = np.array([9, 24, 16, 13, 9, 16, 7, 13, 14, 12])
        np.testing.assert_equal(ads.solus.astype(int).sum().values, rep_sum)
        i_sum = np.array(
            [4, 8, 4, 5, 6, 6, 8, 5, 6, 6, 3, 2, 7, 6, 5, 4, 5, 6, 5, 4, 6, 6, 7, 3, 6]
        )
        np.testing.assert_equal(ads.solus.astype(int).sum(axis=1).values, i_sum)
        # ------------------------#
        #           # Multi Core #
        # ------------------------#
        np.random.seed(10)
        ads = adbscan.ADBSCAN(0.03, 3, reps=10, keep_solus=True, n_jobs=-1)
        with pytest.warns(
            UserWarning,
            match=(
                "Multi-core implementation only works on relabelling solutions. "
                "Execution of draws is still sequential."
            ),
        ):
            _ = ads.fit(self.db, xy=["x", "y"])
        # Params
        assert ads.n_jobs == -1
        # Labels
        np.testing.assert_equal(ads.labels_, self.lbls)
        # Votes
        votes = pandas.DataFrame({"lbls": self.lbls, "pct": self.pcts})
        np.testing.assert_equal(ads.votes["lbls"].values, self.lbls)
        np.testing.assert_almost_equal(ads.votes["pct"].values, votes["pct"].values)
        # Solus (only testing the sums as there're too many values)
        np.testing.assert_equal(ads.solus.astype(int).sum().sum(), 133)
        rep_sum = np.array([9, 24, 16, 13, 9, 16, 7, 13, 14, 12])
        np.testing.assert_equal(ads.solus.astype(int).sum().values, rep_sum)
        i_sum = np.array(
            [4, 8, 4, 5, 6, 6, 8, 5, 6, 6, 3, 2, 7, 6, 5, 4, 5, 6, 5, 4, 6, 6, 7, 3, 6]
        )
        np.testing.assert_equal(ads.solus.astype(int).sum(axis=1).values, i_sum)


class TestRemapLBLS:
    def setup_method(self):
        self.db = pandas.DataFrame({"X": [0, 0.1, 4, 6, 5], "Y": [0, 0.2, 5, 7, 5]})
        self.solus = pandas.DataFrame(
            {
                "rep-00": [0, 0, 7, 7, -1],
                "rep-01": [4, 4, -1, 6, 6],
                "rep-02": [5, 5, 8, 8, 8],
            }
        )

    def test_remap_lbls(self):
        vals = np.array([[0, 0, 0], [0, 0, 0], [7, -1, 7], [7, 7, 7], [-1, 7, 7]])
        # ------------------------#
        #           # Single Core #
        # ------------------------#
        lbls = adbscan.remap_lbls(self.solus, self.db)
        # Column names
        np.testing.assert_equal(self.solus.columns.to_numpy(), lbls.columns.to_numpy())
        # Values
        np.testing.assert_equal(lbls.values, vals)
        # ------------------------#
        #            # Multi Core #
        # ------------------------#
        lbls = adbscan.remap_lbls(self.solus, self.db, n_jobs=-1)
        # Column names
        np.testing.assert_equal(self.solus.columns.to_numpy(), lbls.columns.to_numpy())
        # Values
        np.testing.assert_equal(lbls.values, vals)


class TestEnsemble:
    def setup_method(self):
        self.db = pandas.DataFrame(
            {"X": [0, 0.1, 4, 6, 5], "Y": [0, 0.2, 5, 7, 5]}
        ).rename(lambda i: "i_" + str(i))
        solus = pandas.DataFrame(
            {
                "rep-00": [0, 0, 7, 7, -1],
                "rep-01": [4, 4, -1, 6, 6],
                "rep-02": [5, 5, 8, 8, 8],
            }
        ).rename(lambda i: "i_" + str(i))
        self.solus_relabelled = adbscan.remap_lbls(solus, self.db)

    def test_ensemble(self):
        vals = np.array(
            [
                [0.0, 1.0],
                [0.0, 1.0],
                [7.0, 0.6666666666666666],
                [7.0, 1.0],
                [7.0, 0.6666666666666666],
            ]
        )
        # ------------------------#
        ensemble_solu = adbscan.ensemble(self.solus_relabelled)
        # Column names
        np.testing.assert_equal(ensemble_solu.columns.values.tolist(), ["lbls", "pct"])
        # Index
        # Values
        np.testing.assert_almost_equal(ensemble_solu.values, vals)


class TestGetClusterBoundary:
    def setup_method(self):
        np.random.seed(10)
        self.db = pandas.DataFrame(
            {"x": np.random.random(25), "y": np.random.random(25)}
        )
        self.lbls = np.array(
            [
                "-1",
                "-1",
                "-1",
                "0",
                "-1",
                "-1",
                "-1",
                "0",
                "-1",
                "-1",
                "-1",
                "-1",
                "-1",
                "-1",
                "0",
                "0",
                "0",
                "-1",
                "0",
                "-1",
                "0",
                "-1",
                "-1",
                "-1",
                "-1",
            ],
            dtype=object,
        )
        self.pcts = np.array(
            [
                0.7,
                0.5,
                0.7,
                1.0,
                0.7,
                0.7,
                0.5,
                1.0,
                0.7,
                0.7,
                0.6,
                0.6,
                0.6,
                0.7,
                1.0,
                0.9,
                1.0,
                0.7,
                1.0,
                0.7,
                0.9,
                0.7,
                0.8,
                0.6,
                0.7,
            ]
        )
        np.random.seed(10)
        ads = adbscan.ADBSCAN(0.03, 3, reps=10, keep_solus=True)
        _ = ads.fit(self.db, xy=["x", "y"])
        self.labels = pandas.Series(ads.labels_, index=self.db.index)

    @pytest.mark.skipif(sklearn.__version__ == "1.3.0", reason="sklearn regression")
    def test_get_cluster_boundary(self):
        # ------------------------#
        #           # Single Core #
        # ------------------------#
        polys = adbscan.get_cluster_boundary(self.labels, self.db, xy=["x", "y"])
        wkt = (
            "POLYGON ((0.7217553174317995 0.8192869956700687, 0.7605307121989587 "
            "0.9086488808086682, 0.9177741225129434 0.8568503024577332, "
            "0.8126209616521135 0.6262871483113925, 0.6125260668293881 "
            "0.5475861559192435, 0.5425443680112613 0.7546476915298572, "
            "0.7217553174317995 0.8192869956700687))"
        )
        assert polys.iloc[0].wkt == wkt

        # ------------------------#
        #           # Multi Core #
        # ------------------------#
        polys = adbscan.get_cluster_boundary(
            self.labels, self.db, xy=["x", "y"], n_jobs=-1
        )
        wkt = (
            "POLYGON ((0.7217553174317995 0.8192869956700687, 0.7605307121989587 "
            "0.9086488808086682, 0.9177741225129434 0.8568503024577332, "
            "0.8126209616521135 0.6262871483113925, 0.6125260668293881 "
            "0.5475861559192435, 0.5425443680112613 0.7546476915298572, "
            "0.7217553174317995 0.8192869956700687))"
        )
        assert polys.iloc[0].wkt == wkt



import numpy
import esda
from esda.significance import calculate_significance
import pytest
from libpysal.weights import Voronoi

numpy.random.seed(2478879)
coordinates = numpy.random.random(size=(800, 2))
x = numpy.random.normal(size=(800,))
w = Voronoi(coordinates, clip="bounding_box")
w.transform = "r"
stat = esda.Moran_Local(x, w, permutations=19)

@pytest.mark.parametrize("alternative", ["two-sided", "directed", "lesser", "greater", "folded"])
def test_execution_and_range(alternative):
    out = calculate_significance(stat.Is, stat.rlisas, alternative=alternative)
    assert (out > 0).all() & (out <= 1).all(), f'p-value out of bounds for method {alternative}'
    if alternative == 'directed':
        assert out.max() <= .5, f"max p-value is too large for method {alternative}"
    else:
        assert out.max() >= .5, f"max p-value is too small for method {alternative}"
    

def test_alternative_relationships():
    two_sided = calculate_significance(
        stat.Is, 
        stat.rlisas, 
        alternative="two-sided"
    )
    directed = calculate_significance(
        stat.Is, 
        stat.rlisas, 
        alternative="directed"
    )
    lesser = calculate_significance(
        stat.Is, 
        stat.rlisas, 
        alternative="lesser"
    )
    greater = calculate_significance(
        stat.Is, 
        stat.rlisas, 
        alternative="greater"
    )
    folded = calculate_significance(
        stat.Is, 
        stat.rlisas, 
        alternative="folded"
    )

    numpy.testing.assert_allclose(lesser + greater,
                                  numpy.ones_like(lesser) + (1/(stat.permutations + 1)),
        err_msg = "greater p-value should be complement of lesser"
    )
    assert (directed <= two_sided).all(), "directed is bigger than two_sided and should not be"
    one_or_the_other = (directed == lesser) | (directed == greater)
    assert one_or_the_other.all(), "some directed p-value is neither the greater nor lesser p-value"
    assert (two_sided < folded).mean() < (directed < folded).mean(), (
	"Directed p-values should tend to be much smaller than two_sided p-values or folded p-values."
    )
